// -*- C++ -*-
/*!
 * @file  LRFCalibration.cpp
 * @brief Calibratin of position and orientation of laser range finder based on a set of corresponding points
 * @date $Date$
 *
 * $Id$
 */

#include "LRFCalibration.h"

// Module specification
// <rtc-template block="module_spec">
static const char* lrfcalibration_spec[] =
  {
    "implementation_id", "LRFCalibration",
    "type_name",         "LRFCalibration",
    "description",       "Calibratin of position and orientation of laser range finder based on a set of corresponding points",
    "version",           "1.0.0",
    "vendor",            "TakeshiSasaki",
    "category",          "generic",
    "activity_type",     "DataFlowComponent",
    "max_instance",      "10",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    "conf.default.data_num", "0",

    ""
  };
// </rtc-template>

LRFCalibration::LRFCalibration(RTC::Manager* manager)
  : RTC::DataFlowComponentBase(manager),
    // <rtc-template block="initializer">
    m_GlobalXIn("GlobalX", m_GlobalX),
    m_GlobalYIn("GlobalY", m_GlobalY),
    m_LocalXIn("LocalX", m_LocalX),
    m_LocalYIn("LocalY", m_LocalY),
    m_TxOut("Tx", m_Tx),
    m_TyOut("Ty", m_Ty),
    m_ThetaOut("Theta", m_Theta),
    
    // </rtc-template>
	dummy(0)
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  registerInPort("GlobalX", m_GlobalXIn);
  registerInPort("GlobalY", m_GlobalYIn);
  registerInPort("LocalX", m_LocalXIn);
  registerInPort("LocalY", m_LocalYIn);
  
  // Set OutPort buffer
  registerOutPort("Tx", m_TxOut);
  registerOutPort("Ty", m_TyOut);
  registerOutPort("Theta", m_ThetaOut);
  
  // Set service provider to Ports
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  
  // </rtc-template>

}

LRFCalibration::~LRFCalibration()
{
}



RTC::ReturnCode_t LRFCalibration::onInitialize()
{
  // <rtc-template block="bind_config">
  // Bind variables and configuration variable
  bindParameter("data_num", m_data_num, "0");
  
  // </rtc-template>
  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t LRFCalibration::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t LRFCalibration::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t LRFCalibration::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t LRFCalibration::onActivated(RTC::UniqueId ec_id)
{
  m_LocalXIn.read();
  m_LocalYIn.read();
  m_GlobalXIn.read();
  m_GlobalYIn.read();
  std::cerr << "--Calibration for laser range finder--" << std::endl; 
  std::cerr << "Set data_num to define how many calibration points are used." << std::endl;
  data_size = 0;

  return RTC::RTC_OK;
}


/*
RTC::ReturnCode_t LRFCalibration::onDeactivated(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t LRFCalibration::onExecute(RTC::UniqueId ec_id)
{
  static int count = 0; //number of calibration points obtained
  double gx=0, gy=0, lx=0, ly=0; //corresponding points (global and local coordinates)

  static double sum_gx=0, sum_gy=0, sum_lx=0, sum_ly=0;
  static double c1=0, c2=0;
  double atanX, atanY;

  double tx, ty, theta; //result of calibration - translation and rotation angle

  if(m_data_num == 0){
    data_size = 0;
    return RTC::RTC_OK;
  }else if(m_data_num < 2){
    std::cerr << "Bad number of points (data_num)" << std::endl;
    data_size = 0;
    return RTC::RTC_OK;
  }else if(data_size != m_data_num){ //first time or change m_data_num during calibration -> initialize data
    data_size = m_data_num;
    count = 0;
    sum_gx = 0;
    sum_gy = 0;
    sum_lx = 0;
    sum_ly = 0;
    c1 = 0;
    c2 = 0;
    std::cerr << "Number of points (data_num) is set to " << data_size << std::endl;
    std::cerr << "Initialization ... done." << std::endl;
  }

  if(m_GlobalXIn.isNew() && m_GlobalYIn.isNew() && m_LocalXIn.isNew() && m_LocalYIn.isNew()){  //new data comes
    // Read data from InPort
    m_GlobalXIn.read();
    m_GlobalYIn.read();
    m_LocalXIn.read();
    m_LocalYIn.read();

    gx = m_GlobalX.data;
    gy = m_GlobalY.data;
    lx = m_LocalX.data;
    ly = m_LocalY.data;

    std::cout << "Calibration object -" << count << "- found at (" << gx << ", " << gy << "), (" << lx << ", " << ly << ")." << std::endl;

    sum_gx += gx;
    sum_gy += gy;
    sum_lx += lx;
    sum_ly += ly;
    c1 += gx*ly - gy*lx;
    c2 += gx*lx + gy*ly;

    count++;

    // Enough data... output
    if(count == data_size){
      std::cout << "All objects done..." << std::endl;
      atanX = c2*data_size-sum_gx*sum_lx-sum_gy*sum_ly;
      atanY = -c1*data_size+sum_gx*sum_ly-sum_gy*sum_lx;

      if((atanX == 0) && (atanY == 0)){
        std::cerr << "Input data is not enough or invalid" << std::endl;
        //Initialize data
        count = 0;
        sum_gx = 0;
        sum_gy = 0;
        sum_lx = 0;
        sum_ly = 0;
        c1 = 0;
        c2 = 0;
        return RTC::RTC_OK;
      }

      //calibration
      theta = std::atan2(atanY, atanX);
      tx = (sum_gx - std::cos(theta)*sum_lx + std::sin(theta)*sum_ly) / data_size;
      ty = (sum_gy - std::sin(theta)*sum_lx - std::cos(theta)*sum_ly) / data_size;

      //********** output *****************************//
      std::cout << "Tx:" << tx << " Ty:" << ty << " theta:" << theta << std::endl;
      m_Tx.data = tx;
      m_Ty.data = ty;
      m_Theta.data = theta;
      m_TxOut.write();
      m_TyOut.write();
      m_ThetaOut.write();

      //Initialize data
      count = 0;
      sum_gx = 0;
      sum_gy = 0;
      sum_lx = 0;
      sum_ly = 0;
      c1 = 0;
      c2 = 0;
    }

    
  }


  return RTC::RTC_OK;
}


/*
RTC::ReturnCode_t LRFCalibration::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t LRFCalibration::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t LRFCalibration::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t LRFCalibration::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t LRFCalibration::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/



extern "C"
{
 
  void LRFCalibrationInit(RTC::Manager* manager)
  {
    RTC::Properties profile(lrfcalibration_spec);
    manager->registerFactory(profile,
                             RTC::Create<LRFCalibration>,
                             RTC::Delete<LRFCalibration>);
  }
  
};


