// -*- C++ -*-
/*!
 * @file  LRFComponent.cpp
 * @brief Laser Range Finder Component
 * @date $Date$
 *
 * $Id$
 */

#include "LRFComponent.h"

// Module specification
// <rtc-template block="module_spec">
static const char* lrfcomponent_spec[] =
  {
    "implementation_id", "LRFComponent",
    "type_name",         "LRFComponent",
    "description",       "Laser Range Finder Component",
    "version",           "1.0.0",
    "vendor",            "TakeshiSasaki",
    "category",          "generic",
    "activity_type",     "DataFlowComponent",
    "max_instance",      "10",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    "conf.default.DeviceName", "/dev/ttyACM0",
    "conf.default.ScanRate", "19200",
    "conf.default.ScanStart", "0",
    "conf.default.ScanEnd", "768",
    "conf.default.ScanStep", "1",

    ""
  };
// </rtc-template>

LRFComponent::LRFComponent(RTC::Manager* manager)
  : RTC::DataFlowComponentBase(manager),
    // <rtc-template block="initializer">
    m_ScanDataOut("ScanData", m_ScanData),
    m_StartPointOut("StartPoint", m_StartPoint),
    m_EndPointOut("EndPoint", m_EndPoint),
    m_LRFInfoPort("LRFInfo"),
    // </rtc-template>
	dummy(0)
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  
  // Set OutPort buffer
  registerOutPort("ScanData", m_ScanDataOut);
  registerOutPort("StartPoint", m_StartPointOut);
  registerOutPort("EndPoint", m_EndPointOut);
  
  // Set service provider to Ports
  m_LRFInfoPort.registerProvider("InfoPort", "LRFInfo", m_InfoPort);
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  registerPort(m_LRFInfoPort);
  
  // </rtc-template>
  m_DeviceName = new char[50];
}

LRFComponent::~LRFComponent()
{
}



RTC::ReturnCode_t LRFComponent::onInitialize()
{
  // <rtc-template block="bind_config">
  // Bind variables and configuration variable
  bindParameter("DeviceName", m_DeviceName, "/dev/ttyACM0");
  bindParameter("ScanRate", m_ScanRate, "19200");
  bindParameter("ScanStart", m_ScanStart, "0");
  bindParameter("ScanEnd", m_ScanEnd, "768");
  bindParameter("ScanStep", m_ScanStep, "1");
  
  // </rtc-template>
  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t LRFComponent::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t LRFComponent::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t LRFComponent::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t LRFComponent::onActivated(RTC::UniqueId ec_id)
{
  m_InfoPort.laser.initialize(m_DeviceName,m_ScanRate);

  return RTC::RTC_OK;
}



RTC::ReturnCode_t LRFComponent::onDeactivated(RTC::UniqueId ec_id)
{
  m_InfoPort.laser.closeSerial();

  return RTC::RTC_OK;
}



RTC::ReturnCode_t LRFComponent::onExecute(RTC::UniqueId ec_id)
{
  int i;
  int st = m_ScanStart;
  int ed = m_ScanEnd;
  int stp = m_ScanStep;
  int check_result = 0;
  int iteration = (ed-st)/stp + 1;

  if((st < 0) || (st > m_InfoPort.laser.getMaxStep())){
    std::cerr << "invalid parameter: ScanStart" << std::endl;
    check_result = 1;
  }
  if((ed < 0) || (ed > m_InfoPort.laser.getMaxStep()) || (ed < st)){
    std::cerr << "invalid parameter: ScanEnd" << std::endl;
    check_result = 1;
  }
  if(stp < 0){
    std::cerr << "invalid parameter: ScanStep" << std::endl;
    check_result = 1;
  }

  if(check_result == 0){
    if(m_InfoPort.laser.makeScan(st,ed,stp) == 0){
      m_ScanData.data.length(iteration);

      for(i=0;i<iteration;i++){
        m_ScanData.data[i] = m_InfoPort.laser.data[i];
      }

      m_StartPoint.data = st;
      m_EndPoint.data = ed;

      m_ScanDataOut.write();
      m_StartPointOut.write();
      m_EndPointOut.write();
    }else{
      std::cerr << "scan failed" << std::endl;
    }
  }

  return RTC::RTC_OK;
}


/*
RTC::ReturnCode_t LRFComponent::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t LRFComponent::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t LRFComponent::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t LRFComponent::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t LRFComponent::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/



extern "C"
{
 
  void LRFComponentInit(RTC::Manager* manager)
  {
    RTC::Properties profile(lrfcomponent_spec);
    manager->registerFactory(profile,
                             RTC::Create<LRFComponent>,
                             RTC::Delete<LRFComponent>);
  }
  
};


