// -*- C++ -*-
/*!
 * @file  SimpleTracker.cpp
 * @brief Single object tracker
 * @date $Date$
 *
 * $Id$
 */

#include "SimpleTracker.h"

// Module specification
// <rtc-template block="module_spec">
static const char* simpletracker_spec[] =
  {
    "implementation_id", "SimpleTracker",
    "type_name",         "SimpleTracker",
    "description",       "Single object tracker",
    "version",           "1.0.0",
    "vendor",            "TakeshiSasaki",
    "category",          "generic",
    "activity_type",     "DataFlowComponent",
    "max_instance",      "10",
    "language",          "C++",
    "lang_type",         "compile",
    ""
  };
// </rtc-template>

SimpleTracker::SimpleTracker(RTC::Manager* manager)
  : RTC::DataFlowComponentBase(manager),
    // <rtc-template block="initializer">
    m_ScanDataIn("ScanData", m_ScanData),
    m_StartPointIn("StartPoint", m_StartPoint),
    m_EndPointIn("EndPoint", m_EndPoint),
    m_XOut("X", m_X),
    m_YOut("Y", m_Y),
    m_LRFInfoPort("LRFInfo"),
    // </rtc-template>
	dummy(0)
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  registerInPort("ScanData", m_ScanDataIn);
  registerInPort("StartPoint", m_StartPointIn);
  registerInPort("EndPoint", m_EndPointIn);
  
  // Set OutPort buffer
  registerOutPort("X", m_XOut);
  registerOutPort("Y", m_YOut);
  
  // Set service provider to Ports
  
  // Set service consumers to Ports
  m_LRFInfoPort.registerConsumer("InfoPort", "LRFInfo", m_InfoPort);
  
  // Set CORBA Service Ports
  registerPort(m_LRFInfoPort);
  
  // </rtc-template>

  scandata = NULL;
  background = NULL;
}

SimpleTracker::~SimpleTracker()
{
  if(scandata != NULL){
    delete [] scandata;
    scandata = NULL;
  }
  if(background != NULL){
    delete [] background;
    background = NULL;
  }
}


/*
RTC::ReturnCode_t SimpleTracker::onInitialize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t SimpleTracker::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t SimpleTracker::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t SimpleTracker::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t SimpleTracker::onActivated(RTC::UniqueId ec_id)
{
  m_ScanDataIn.read();
  m_StartPointIn.read();
  m_EndPointIn.read();

  double PI = std::acos(-1);

  try{
    angular_resolution = ((double)(m_InfoPort-> getMaxAngle()) - (m_InfoPort-> getMinAngle()))*(PI/180)/(m_InfoPort-> getMaxStep());
    min_angle = (m_InfoPort-> getMinAngle())*(PI/180);
  }catch(...){
    std::cerr << "Error in service port." << std::endl;
    return RTC::RTC_ERROR;
  }

  data_length=-1;
  start_point=-1;
  end_point=-1;

  return RTC::RTC_OK;
}



RTC::ReturnCode_t SimpleTracker::onDeactivated(RTC::UniqueId ec_id)
{
  if(scandata != NULL){
    delete [] scandata;
    scandata = NULL;
  }
  if(background != NULL){
    delete [] background;
    background = NULL;
  }
  return RTC::RTC_OK;
}



RTC::ReturnCode_t SimpleTracker::onExecute(RTC::UniqueId ec_id)
{
  int i,j,k; //for loop
  static int back_count = 0; //counter for learning background

  int step;
  double th;

  //foreground data
  int fg_count = 0; //how many foreground points found
  double fg_x; //x coordninate of the new foreground point
  double fg_y; //y coordninate of the new foreground point
  int fg_group; //group number of the new foreground point
  int fg_num;
  std::vector<double> fg_array_x; //list of foreground points: x
  std::vector<double> fg_array_y; //y
  std::vector<int> fg_array_g; //group
  std::vector<int> fg_array_n; //number of points in the cluseter

  int group_tmp; //tmp
  int max_num; //maximum number of points in the cluseter
  double x_cl=0, y_cl=0; //position of the cluster center

  if(m_StartPointIn.isNew() && m_EndPointIn.isNew() && m_ScanDataIn.isNew()){ //new data comes
    m_StartPointIn.read();
    m_EndPointIn.read();
    m_ScanDataIn.read();

    //********** error check and initialization *****************************//
    if((start_point != -1) && (start_point != m_StartPoint.data)){ //error check
      std::cerr << "Start point cannot be changed during tracking." << std::endl;
      return RTC::RTC_OK;
    }else if(start_point == -1){ //set start point
      start_point = m_StartPoint.data;
      std::cerr << "Start point set to " << start_point << std::endl;
    }

    if((end_point != -1) && (end_point != m_EndPoint.data)){ //error check
      std::cerr << "End point cannot be changed during tracking." << std::endl;
      return RTC::RTC_OK;
    }else if(end_point == -1){ //set end point
      end_point = m_EndPoint.data;
      std::cerr << "End point set to " << end_point << std::endl;
    }

    if((data_length != -1) && (data_length != (int)m_ScanData.data.length())){ //error check
      std::cerr << "Data length cannot be changed during tracking." << std::endl;
      return RTC::RTC_OK;
    }else if(data_length == -1){ //set data length and initialization of background
      data_length = (int)m_ScanData.data.length();
      std::cerr << "Data length set to " << data_length << std::endl;

      try{
        scandata = new int[data_length];
        background = new int[data_length];
      }catch(std::bad_alloc){
        std::cerr << "Bad alloc exception" << std::endl;
        return RTC::RTC_ERROR;
      }

      for(i=0;i<data_length;i++){
        background[i] = 0;
      }
      back_count = 0;
    }
    //********** end error check and initialization *****************************//

    for(i=0;i<data_length;i++){
      scandata[i] = m_ScanData.data[i];
    }

    //********** get background *****************************//
    if(back_count < 20){
      if(back_count == 0){std::cout << "Learning background...";}
      // --- Get initial background -----
      // the background is set to the filtered value of 20 scans
      for(i=0;i<data_length;i++){
        if(scandata[i] > 20){ //valid data
          if(background[i] > 0){background[i] = (9*background[i] + scandata[i])/10;}
          else{background[i] = scandata[i];}
        }
      }
      back_count++;
      if(back_count == 20){std::cout << "done" << std::endl;}
      return RTC::RTC_OK;
    }
    //********** end get background *****************************//

    //********** main processing *****************************//
    step = (end_point - start_point) / (data_length -1);
    th = min_angle + start_point * angular_resolution;

    for(i=0;i<data_length;i++){
      if(scandata[i] > 20){ //valid data
        if(std::abs(background[i] - scandata[i]) > 200){  // find points in foreground
          // condition: < 3.0 m
//          if(scandata[i] < 3000){
          //********** clustering (for forground points) *****************************//
            //store data
            fg_x = (double)scandata[i]*std::cos(th);
            fg_y = (double)scandata[i]*std::sin(th);
            fg_group = fg_count;
            fg_num = 1;

            //search cluster array
            for(j=0;j<fg_count;j++){
              if((fg_group != fg_array_g[j]) && (std::sqrt((fg_x-fg_array_x[j])*(fg_x-fg_array_x[j]) + (fg_y-fg_array_y[j])*(fg_y-fg_array_y[j])) < 100)){ //small distance -> same cluster
                if(fg_num == 1){ //first time
                  //update number of points
                  fg_array_n[fg_array_g[j]]++;
                  fg_num = -1;
                  //update group number
                  fg_group = fg_array_g[j];
                }else{
                  //update number of points
                  fg_array_n[fg_group] += fg_array_n[fg_array_g[j]];
                  fg_array_n[fg_array_g[j]] = -1;
                  group_tmp = fg_array_g[j];
                  //update group number
                  for(k=0;k<fg_count;k++){
                   if(fg_array_g[k] == group_tmp){
                     fg_array_g[k] = fg_group;
                   }
                  }
                }
              }
            } //end for(j)

            fg_array_x.push_back(fg_x);
            fg_array_y.push_back(fg_y);
            fg_array_g.push_back(fg_group);
            fg_array_n.push_back(fg_num);

            fg_count++;
//          }
        }else{ // the point is background
          //********** update background (for background points) *****************************//
          if(background[i] > 0){ 
            background[i] = (9*background[i] + scandata[i])/10;
          }else{ //no background -> get background
            background[i] = scandata[i];
          }
        }
      }

      th += angular_resolution*step;
    } //end for(i)


    //********** find largest cluser and calculate its center *****************************//
    max_num = -1;
    for(i=0;i<fg_count;i++){
      if(fg_array_n[i] > max_num){
        max_num = fg_array_n[i];
        group_tmp = i;
      }
    }

    if(max_num > 5){  //find object
      for(i=0;i<fg_count;i++){
        if(fg_array_g[i] == group_tmp){
          x_cl += fg_array_x[i];
          y_cl += fg_array_y[i];
        }
       }
       x_cl /= max_num;
       y_cl /= max_num;

      //********** output *****************************//
      //convert millimeter to meter
      x_cl /= 1000;
      y_cl /= 1000;

      std::cout << "x:" << x_cl << " y:" << y_cl << std::endl;
      m_X.data = x_cl;
      m_Y.data = y_cl;
      m_XOut.write();
      m_YOut.write();
    }
    //********** end main processing *****************************//

  }//end if(isNew())


  return RTC::RTC_OK;
}



RTC::ReturnCode_t SimpleTracker::onAborting(RTC::UniqueId ec_id)
{
  if(scandata != NULL){
    delete [] scandata;
    scandata = NULL;
  }
  if(background != NULL){
    delete [] background;
    background = NULL;
  }
  return RTC::RTC_OK;
}


/*
RTC::ReturnCode_t SimpleTracker::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t SimpleTracker::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t SimpleTracker::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t SimpleTracker::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/



extern "C"
{
 
  void SimpleTrackerInit(RTC::Manager* manager)
  {
    RTC::Properties profile(simpletracker_spec);
    manager->registerFactory(profile,
                             RTC::Create<SimpleTracker>,
                             RTC::Delete<SimpleTracker>);
  }
  
};


